/*
 * file:   man1d_acc.c
 * author: Hinke Osinga - University of Bristol
 * date:   November 1997
 */

/*
 * Norm(int dim, double *p, double *q, double *qlift, Manifold *proj)
 * Compute_Angle(double *u0, double *u1, double *u2, int dim, 
 *               Manifold *proj)
 * Check_Point(double *p1, double *p2, double *p3, int dim, 
 *	       double *angle, double *dist, double *plift, Manifold *proj)
 * Bisect_Section(double *pleft, double *pright, double *p, int dim, 
 * 		  double delta, double fudge, double *pnew, double *invpnew, 
 * 		  struct Prop_DataS *iter_cntl, int *sw)
 * Map_Points(List inv_item, double *p, int dim, double delta, 
 *	      struct Prop_DataS *iter_cntl, int *sw, int *status)
 * Find_Point(double *pnew, double *invpnew, double *p, double *invp, 
 *	      int dim, List inv_item, double delta, double fudge, 
 *	      struct Prop_DataS *iter_cntl, int *sw, int *status)
 */


#include <stdio.h>
#include <math.h>

#include "complib.h"
#include "utilities.h"
#include "math_utils.h"
#include "manifold.h"
#include "man1d_funct.h"
#include "man1d_list.h"


double Norm(int dim, double *p, double *q, double *qlift,
		    Manifold *proj)
{
  double  dist_square;

  dist_square = lift(dim, p, q, qlift, proj);
  return( sqrt(dist_square) );
}

double Compute_Angle(double *u0, double *u1, double *u2, int dim,
		     Manifold *proj)
{
  int     i;
  double  norm_u1u2, norm_u1u0, norm_bar;
  double  *u1minu2 = dvector(0, dim-1);
  double  *u1minu0 = dvector(0, dim-1);
  double  *u0_bar  = dvector(0, dim-1);
  double  *baru0   = dvector(0, dim-1);

  /* lift u0 and u2 with respect to u1; compute norms u1-u2 and u1-u0 */
  norm_u1u2 = lift(dim, u1, u2, u1minu2, proj);
  norm_u1u2 = sqrt(norm_u1u2);
  norm_u1u0 = lift(dim, u1, u0, u1minu0, proj);
  norm_u1u0 = sqrt(norm_u1u0);
  /* compute u1-u2 and u1-u0, with lifted version of u0 and u2 */
  for (i = 0; i < dim; i++){
    u1minu2[i] = u1[i] - u1minu2[i];
    u1minu0[i] = u1[i] - u1minu0[i];
  }

  /* in covering space: compute u0_bar and the norm of u0_bar - u0 */
  norm_bar = 0;
  for (i = 0; i < dim; i++){
    u0_bar[i] = u1[i] + norm_u1u0 * u1minu2[i] / norm_u1u2;
    baru0[i] = u0_bar[i] - u0[i];
    norm_bar += baru0[i] * baru0[i];
  }
  norm_bar = sqrt(norm_bar);

  free_dvector(u1minu2, 0, dim-1);
  free_dvector(u1minu0, 0, dim-1);
  free_dvector(u0_bar,  0, dim-1);
  free_dvector(baru0,   0, dim-1);

  return (norm_bar / norm_u1u0);
}

void Check_Point(double *p1, double *p2, double *p3, int dim, 
		 double *angle, double *dist, double *plift, 
		 Manifold *proj)
{
  *angle = Compute_Angle(p1, p2, p3, dim, proj);
  *dist = Norm(dim, p2, p3, plift, proj);
}

int Bisect_Section(double *pleft, double *pright, double *p, int dim, 
		   double delta, double fudge, double *pnew, 
		   double *invpnew, struct Prop_DataS *iter_cntl, int *sw)
{
  int      i, status = 0;
  double   tau, norm, fudge_left, fudge_right;
  double   *pl = dvector(0, dim);
  double   *pr = dvector(0, dim);
  double   *pm = dvector(0, dim);
  double   *fpl = dvector(0, dim);
  double   *fpr = dvector(0, dim);
  double   *fpm = dvector(0, dim);

  fudge_right = fudge * delta;
  fudge_left = delta - fudge_right;
  fudge_right += delta;

  for (i = 0; i <= dim; i++){
    pl[i] = pleft[i];
    pr[i] = pright[i];
  }

  norm = Norm(dim, pl, pr, pr, iter_cntl->manifold);
  status = F(fpl, pl, iter_cntl, sw);
  if (status >= 0){
    tau = Norm(dim, p, fpl, iter_cntl->workspace, iter_cntl->manifold);
    if ((tau > fudge_left) && (tau < fudge_right)){
      for (i = 0; i <= dim; i++){
	pnew[i] = fpl[i];
	invpnew[i] = pl[i];
      }
      free_dvector(pl, 0, dim);
      free_dvector(pr, 0, dim);
      free_dvector(pm, 0, dim);
      free_dvector(fpl, 0, dim);
      free_dvector(fpr, 0, dim);
      free_dvector(fpm, 0, dim);

      return status;
    }
    status = F(fpr, pr, iter_cntl, sw);
    if (status >= 0){
      tau = Norm(dim, p, fpr, iter_cntl->workspace, iter_cntl->manifold);
      if ((tau > fudge_left) && (tau < fudge_right)){
	for (i = 0; i <= dim; i++){
	  pnew[i] = fpr[i];
	  invpnew[i] = pr[i];
	}
	free_dvector(pl, 0, dim);
	free_dvector(pr, 0, dim);
	free_dvector(pm, 0, dim);
	free_dvector(fpl, 0, dim);
	free_dvector(fpr, 0, dim);
	free_dvector(fpm, 0, dim);

	return status;
      }
    }
  }

  while (((tau < fudge_left) || (tau > fudge_right)) && 
	 (norm > 1.0e-16) && (status >= 0)){
    for (i = 0; i <= dim; i++)
      pm[i] = (pl[i] + pr[i]) / 2;
    status = F(fpm, pm, iter_cntl, sw);
    if (status >= 0){
      tau = Norm(dim, p, fpm, iter_cntl->workspace, iter_cntl->manifold);
      if (tau > fudge_right)
	for (i = 0; i <= dim; i++)
	  pr[i] = pm[i];
      else
	for (i = 0; i <= dim; i++)
	  pl[i] = pm[i];
      norm = euclid_dist(dim, pl, pr);
    }
  }
  
  if (status >= 0){
    if ((tau < fudge_left) || (tau > fudge_right)){
      system_mess_proc(1, "Preimages are equal, yet no point found:");
      system_mess_proc(1, "ERROR: manifold stretches too much!");
      status = -1;
    }

    if (status >= 0){
      for (i = 0; i <= dim; i++){
	pnew[i] = fpm[i];
	invpnew[i] = pm[i];
      }
    }
  }

  free_dvector(pl, 0, dim);
  free_dvector(pr, 0, dim);
  free_dvector(pm, 0, dim);
  free_dvector(fpl, 0, dim);
  free_dvector(fpr, 0, dim);
  free_dvector(fpm, 0, dim);

  return status;
}

List Map_Points(List inv_item, double *p, int dim, double delta, 
		struct Prop_DataS *iter_cntl, int *sw, int *status)
{
  int     i;
  double  dist, *q, *fq;
  List    list_elem;

  q = dvector(0, dim);
  fq = dvector(0, dim);

  /* find interval that is mapped by F about distance delta from p */
  list_elem = inv_item;
  do {
    list_elem = list_elem->next;
    if (list_elem != NULL){
      for (i = 0; i <= dim; i++)
	q[i] = list_elem->data[i];
      *status = F(fq, q, iter_cntl, sw);
      if (*status >= 0)
	dist = Norm(dim, p, fq, iter_cntl->workspace, iter_cntl->manifold);
    }
  } while ((dist < delta) && (*status >= 0) && (list_elem != NULL));

  free_dvector(q, 0, dim);  
  free_dvector(fq, 0, dim);

  if (*status < 0)
    return NULL;
  else{
    if (list_elem == NULL)
      return list_elem;
    else
      return list_elem->prev;
  }
}

List Find_Point(double *pnew, double *invpnew, double *p, double *invp, 
		int dim, List inv_item, double delta, double fudge, 
		struct Prop_DataS *iter_cntl, int *sw, int *status)
{
  int      i, st = 0;
  double   *pleft = dvector(0, dim);
  double   *pright = dvector(0, dim);
  List     newinv_item;

  /* Find inverval whose image intersects circle around p 
     with radius delta                                     */
  newinv_item = Map_Points(inv_item, p, dim, delta, iter_cntl, sw, &st);
  *status = st;
  if (newinv_item == NULL)
    return newinv_item;

  /* next do bisection on this interval; start with invp when
     interval starts with inv_item */
  if (newinv_item == inv_item){
    for (i = 0; i <= dim; i++){
      pleft[i] = invp[i];
      pright[i] = newinv_item->next->data[i];
    }
  }
  else{
    for (i = 0; i <= dim; i++){
      pleft[i] = newinv_item->data[i];
      pright[i] = newinv_item->next->data[i];
    }
  }
  *status = Bisect_Section(pleft, pright, p, dim, delta, fudge, 
			   pnew, invpnew, iter_cntl, sw);

  free_dvector(pleft, 0, dim);
  free_dvector(pright, 0, dim);

  if (*status >= 0)
    return newinv_item;
  else
    return NULL;
}
