/*
 * file:   man1d_build.c
 * author: Hinke Osinga - University of Bristol
 * date:   November 1997
 */

/*
 * man1d_build(struct Man1D_DataS *s, int pos, double delta_start, 
 *	       double *params)
 * man1d_iterate(struct Man1D_DataS *s, double *params)
 * man1d_grow(struct Man1D_DataS *s, double *params)
 */


#include <stdio.h>
#include <stdlib.h>

#include "utilities.h"
#include "math_utils.h"
#include "plot.h"
#include "man1d_funct.h"
#include "man1d.h"


int man1d_build(struct Man1D_DataS *s, int pos, double delta_start,
		double *params)
{
  int           var_dim, i, status = 0;
  double        norm;
  double        *ev, *p;
  List          item;

  var_dim = s->v_dim-1;

  status = add_to_mfld(&(s->mfld), s->fixpt, s->fixpt, NULL);
  if (status >= 0){
    	item = s->mfld.head;
    	if (s->arclength < fabs(delta_start)){
      		if (delta_start > 0)
			delta_start = s->arclength;
      		else
			delta_start = -s->arclength;
    	}
    
    	/* copy and compute the norm of the eigenvector */
    	ev = dvector(0, var_dim-1);
    	norm = 0;
    	for (i = 0; i < var_dim; i++){
      		ev[i] = s->evectors[i][pos];
      		norm += ev[i]*ev[i];
    	}
    	norm = sqrt(norm);

    	/* normalize */
    	for (i = 0; i < var_dim; i++)
      	ev[i] /= norm;

    	p = dvector(0, var_dim);

    	for (i = 0; i < var_dim; i++)
      		p[i] = s->fixpt[i] + delta_start * ev[i];
    	/* copy time */
    	p[var_dim] = s->fixpt[var_dim];
    	project(var_dim, p, s->prop_cntl.manifold);
    	status = add_to_mfld(&(s->mfld), p, s->fixpt, item);
    	s->mfld.arclength = fabs(delta_start);
    	point_plot(p, params, s->prop_cntl.sys_color, s->prop_cntl.symbol);
  }

  free_dvector(ev, 0, var_dim-1);
  free_dvector(p, 0, var_dim);

  return(status);
}

/* simple iterate as long as images are close enough together */
int man1d_iterate(struct Man1D_DataS *s, double *params)
{
  int        var_dim, i, status = 0;
  int        sw[3]; 
  double     alpha, delta;
  double     *p, *fp, *pbuur;
  List       item;

  sw[0] = s->explicit_switch;
  sw[1] = s->approx_switch;
  sw[2] = s->jacobian_switch;
  var_dim = s->v_dim-1;

  p = dvector(0, var_dim);
  fp = dvector(0, var_dim);
  pbuur = dvector(0, var_dim);

  item = s->mfld.tail;
  for (i = 0; i <= var_dim; i++){
    pbuur[i] = item->prev->data[i];
    p[i] = item->data[i];
  }

  status = F(fp, p, &(s->prop_cntl), sw);
  if (status >= 0)
    Check_Point(pbuur, p, fp, var_dim, &alpha, &delta, 
		s->prop_cntl.workspace, s->prop_cntl.manifold);
  while ((alpha < s->amax) && (alpha * delta < s->damax) &&
	 (s->mfld.arclength < s->arclength) && 
	 (delta > s->converge) && (status >= 0)){
    point_plot(fp, params, s->prop_cntl.sys_color, s->prop_cntl.symbol);
    status = add_to_mfld(&(s->mfld), fp, p, item);
    s->mfld.arclength += delta;
    item = s->mfld.tail;
    for (i = 0; i <= var_dim; i++){
      pbuur[i] = p[i];
      p[i] = fp[i];
    }
    status = F(fp, p, &(s->prop_cntl), sw);
    if (status >= 0)
      Check_Point(pbuur, p, fp, var_dim, &alpha, &delta, 
		  s->prop_cntl.workspace, s->prop_cntl.manifold);
  }

  printf("finished iterating initial section - arclength = %f\n", 
	 s->mfld.arclength);

  if (delta <= s->converge)
    fprintf(stdout, "Convergence to fixed point.");

  free_dvector(p, 0, var_dim);
  free_dvector(fp, 0, var_dim);
  free_dvector(pbuur, 0, var_dim);

  return status;
}




/* start our algorithm when iterates become too far apart */
int man1d_grow(struct Man1D_DataS *s, double *params)
{
  int        var_dim, i, status = 0;
  int        sw[3]; 
  double     alpha, delta, da, delta_try, half_dmin;
  double     *p, *pnew, *pbuur, *invp, *invpnew;
  List       item, inv_item;

  sw[0] = s->explicit_switch;
  sw[1] = s->approx_switch;
  sw[2] = s->jacobian_switch;
  var_dim = s->v_dim-1;

  p = dvector(0, var_dim);
  pnew = dvector(0, var_dim);
  pbuur = dvector(0, var_dim);
  invp = dvector(0, var_dim);
  invpnew = dvector(0, var_dim);

  item = s->mfld.tail;
  for (i = 0; i <= var_dim; i++){
    pbuur[i] = item->prev->data[i];
    p[i] = item->data[i];
    invp[i] = item->inverse[i];
  }

  delta = Norm(var_dim, pbuur, p, s->prop_cntl.workspace, 
	       s->prop_cntl.manifold);
  inv_item = item->invleft;
  delta_try = delta;
  if (delta_try < s->dmin)
    delta_try = s->dmin;
  half_dmin = s->dmin / 2;
  
  while ((s->mfld.arclength < s->arclength) &&
	 (delta > s->converge) && (status >= 0)){
    inv_item = Find_Point(pnew, invpnew, p, invp, var_dim, inv_item, 
			  delta_try, s->fudge, &(s->prop_cntl), sw, &status);
    if (inv_item == NULL){
      if (status >= 0){
	inv_item = s->mfld.tail;
	delta_try /= 2;
	if (delta_try < half_dmin)
	  delta_try = half_dmin;
	status = F(pnew, p, &(s->prop_cntl), sw);
	if (status >= 0)
	  for (i = 0; i <= var_dim; i++)
	    invpnew[i] = p[i];
      }
    }

    if (status >= 0){
      Check_Point(pbuur, p, pnew, var_dim, &alpha, &delta,
		  s->prop_cntl.workspace, s->prop_cntl.manifold);
      da = delta * alpha;
      if (((alpha < s->amax) && (da < s->damax)) ||
	  (delta < s->dmin)){
	if ((delta < s->dmin) &&
	    ((alpha > s->amax) || (da > s->damax))){	  
	  fprintf(stdout, "delta = %.10f\n", delta);
	  fprintf(stdout, "alpha = %f, delta * alpha = %f; added anyway\n",
		  alpha, da);
	}
      
	/* alpha is acceptable => add point to s->mfld */
	point_plot(pnew, params, s->prop_cntl.sys_color, s->prop_cntl.symbol);
	status = add_to_mfld(&(s->mfld), pnew, invpnew, inv_item);
	s->mfld.arclength += delta;
      
	for (i = 0; i <= var_dim; i++){
	  pbuur[i] = p[i];
	  p[i] = pnew[i];
	  invp[i] = invpnew[i];
	}
	
	if ((alpha < s->amin) && (da < s->damin))
	  /* points are almost collinear => increase delta_try */
	  delta_try *= 2;
      }
      else{
	/* find point closer to p => decrease delta_try */
	inv_item = s->mfld.tail->invleft;
	delta_try /= 2;
	if (delta_try < half_dmin)
	  delta_try = half_dmin;
      }
    }
  }

  if ((status >= 0) && (delta <= s->converge))
    fprintf(stdout, "Convergence to fixed point: ");

  free_dvector(p, 0, var_dim);
  free_dvector(invp, 0, var_dim);
  free_dvector(pnew, 0, var_dim);
  free_dvector(invpnew, 0, var_dim);
  free_dvector(pbuur, 0, var_dim);

  return status;
}
