/*
 * file:   man1d_compute.c
 * author: Hinke Osinga - University of Bristol
 * date:   October 1997
 *
 * edited: James England - University of Bristol
 * date:   November- February 2002
 */

/* 
 * man1d_compute(void)
 * man1ds_compute(void)
 * man1du_compute(void)
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "pm.h"
#include "utilities.h"
#include "math_utils.h"
#include "constants.h"
#include "defaults.h"
#include "complib.h"
#include "iterate.h"
#include "flow.h"
#include "fixed.h"
#include "man1d.h"


void man1du_compute(void)
{
  man1d_compute(2); /* compute unstable manifold */
}

void man1ds_compute(void)
{
  man1d_compute(1); /* compute stable manifold */
}

void man1d_compute(int st_un)
{
  int                 status = 0, num = 0, side_cntr, nside, eval_cntr;
  int                 jac_trans[]={FORW_DIFF, CEN_DIFF, ANALYTIC};
  int           format = *((int *) pm(GET, "Defaults.Precision", NULL));
  int           recording = *((int *) pm (GET, "Defaults.Recording", NULL));
  int                 s_cntr, u_cntr, s1d, u1d;
  int                 i, j, var_dim;
  int                 *extra_iparams, n_extra_iparams;
  double              length_eval, delta_start, *points, *params;
  struct Fixpt_DataS  fp;
  struct Man1D_DataS  s;

  /* need info about fixed points panel */
  status = fp_load(&fp, 0);

  /* load information about the system */
  if (man1d_load(&s) == -1){
    system_mess_proc(1, "man1d_compute: Memory alloc error.\n");
    return;
  }
  else
    var_dim = s.v_dim - 1;

  if (fp.prop_cntl.mapping_flag){
    system_mess_proc(0, "Computing 1D-manifolds of fixed points.\n");
    reset_interrupt();
    iteration_set_up(&(s.prop_cntl));
  }
  else{
    system_mess_proc(1, "Sorry, only works for maps!\n");
    return;
  }

  /* how to compute inverse */
  if ((s.prop_cntl.panel_choice_values[0] > 2) ||
      (s.prop_cntl.panel_choice_values[0] < 0))
    s.prop_cntl.panel_choice_values[0] = 2;
  s.jacobian_switch = jac_trans[s.prop_cntl.panel_choice_values[0]];
  if( !s.prop_cntl.jacobian_flag && (s.jacobian_switch == ANALYTIC))
    s.jacobian_switch = FORW_DIFF;
  
  s.approx_switch = s.prop_cntl.panel_choice_values[1];
  if (s.approx_switch && 
      ( !s.prop_cntl.inverse_flag || s.prop_cntl.explicit_inverse) ) 
    s.approx_switch = FALSE; 

  s.explicit_switch = s.prop_cntl.panel_choice_values[2];
  if (s.explicit_switch && !s.prop_cntl.explicit_inverse)
    s.explicit_switch = FALSE;

  switch (s.side){
  case BOTH :
    side_cntr = 2;
    break;
  case POSITIVE :
    side_cntr = 1;
    break;
  case NEGATIVE :
    side_cntr = 1;
    break;
  default:
    break;
  }

  /* search memory object for any fixed points */
  if (memory_reset_read(s.memory) == 0){
    while (memory_read_next_flow(s.memory, NULL, NULL, NULL,
				 &extra_iparams, &n_extra_iparams) == 0
	   && status >= 0){
      while (memory_read_next_traj(s.memory, &(fp.fp_map_period),
				   NULL, NULL) == 0
	     && status >= 0){
	if (extra_iparams[0] == SADDLE){
	  num++;
	  while (memory_read_next_point(s.memory, &points, &params,
					NULL, NULL, NULL) == 0
		 && status >= 0){
	    /* compute eigenvalues of fixed point */

       
	    for (j = 0; j < s.prop_cntl.parameter_dim; j++)
	      s.prop_cntl.parameters[j] = params[j];
	    for (j = 0; j < s.v_dim; j++){
	      s.fixpt[j] = points[j];
	    }
	    status = man1d_linear(&s, &fp, params);
	    /* check for one-dimensional (un)stable manifolds */
	    s1d = 0;
	    u1d = 0;
	    s_cntr = -1;
	    u_cntr = -1;
	    for (eval_cntr = 0; 
		 eval_cntr < var_dim && status >= 0; eval_cntr++){
	      length_eval = fabs(s.evalues[eval_cntr]);
	      if (length_eval != 1.) {
		if (length_eval > 1.){
		  u1d++;
		  u_cntr = eval_cntr;
		}
		else{
		  s1d++;
		  s_cntr = eval_cntr;
		}
	      }
	    }
	    if (u1d > 1)
	      u_cntr = -1;
	    if (s1d > 1)
	      s_cntr = -1;
	    /* compute 1D (un)stable manifolds only */
	    if ((s1d != 1) && (u1d != 1)){
	      fprintf(stdout, "(");
	      for (i = 0; i < var_dim; i++)
		fprintf(stdout, "%.*lg ", format, s.fixpt[i]);
	      fprintf(stdout, ") has no 1D (un)stable manifolds\n");
	    }
	    while ((s1d == 1) || (u1d == 1)){
	      s.prop_cntl.iter_request = s.map_per;
	      if (u1d == 1 && st_un == 2){
                s1d = 0; /* don't compute stable manifold */
		/* unstable manifold */
		eval_cntr = u_cntr;
		s.man_type = UNSTAB_MAN;
		s.prop_cntl.sys_color = UNSTAB_MAN_COLOR;
		s.prop_cntl.symbol = UNSTAB_MAN_SYMBOL;
		s.prop_cntl.direction = FORWARD;
		s.prop_cntl.time_step = STEP_FORW;
		if (s.evalues[eval_cntr] < 0) {
		  /*   iterating twice if negative */
		  s.prop_cntl.iter_request *= 2;		  
		}     
		if (s.unstabarc > 0){
		  fprintf(stdout, "Computing unstable manifold of (");
		  for (i = 0; i < var_dim; i++)
		    fprintf(stdout, "%.*lg ", format, s.fixpt[i]);
		  fprintf(stdout, "):\n");
		}
		s.arclength = s.unstabarc;
		u1d = -1;
	      }
	      else{
		if (s1d == 1 && st_un == 1){
                  u1d = 0; /* don't compute unstable manifold */
		  /* stable manifold */
		  eval_cntr = s_cntr;
		  s.man_type = STAB_MAN;
		  s.prop_cntl.sys_color = STAB_MAN_COLOR;
		  s.prop_cntl.symbol = STAB_MAN_SYMBOL;
		  s.prop_cntl.direction = BACKWARD;
		  s.prop_cntl.time_step = STEP_BACK;
		  if (s.evalues[eval_cntr] < 0)
		    /*   iterating twice if negative */
		    s.prop_cntl.iter_request *= 2;
		  if (s.stabarc > 0){
		    fprintf(stdout, "Computing stable manifold of (");
		    for (i = 0; i < var_dim; i++)
		      fprintf(stdout, "%.*lg ", format, s.fixpt[i]);
		    fprintf(stdout, "):\n");

		    /* print out info */
		    if (s.explicit_switch && s.method!=2 && s.method!=1)
		      fprintf(stdout, "Using explicit inverse.\n");
		    else if( s.method==1){ /* added by JE */
		      fprintf(stdout,"Using search Circle Method. \n");
		    }
                    else {
		      if (s.approx_switch)
			fprintf(stdout, "Using approximate inverse ");
		      else
			fprintf(stdout, "Using Monte Carlo method ");
		      if (s.jacobian_switch == ANALYTIC)
			fprintf(stdout, "with an analytic jacobian.\n");
		      else if (s.jacobian_switch == FORW_DIFF)
			fprintf(stdout, "with forward differencing.\n");
		      else if (s.jacobian_switch == CEN_DIFF)
			fprintf(stdout, "with central differencing.\n");
		      else
			fprintf(stdout, 
			  "with an error in determining the jacobian.\n");
		    }
                     
		  }
		  s.arclength = s.stabarc;
		  s1d = -1;
		}
	      }

	      if (s.arclength > 0){
		fprintf(stdout, "Ev = %.*lg\n", 
			format, s.evalues[eval_cntr]);
		fprintf(stdout, "  Eigenvector: (");
		for (i = 0; i < var_dim; i++){
		  fprintf(stdout, "%.*lg ",
			  format, s.evectors[i][eval_cntr]);
		}
		fprintf(stdout, ")\n");
		delta_start = s.dinit;

		if (s.evalues[eval_cntr] < 0){ 
		  if(s.max_iter < s.prop_cntl.iter_request){
		    s.max_iter = s.prop_cntl.iter_request; 
		    printf("Neg eigenvalue -> use second iterate. Max iterations must be > %d. Set Max iteration = %d\n"
			   ,s.max_iter, s.max_iter);
		  }	  
		}
		for (nside = 0; nside < side_cntr; nside++){
		  if (s.side == NEGATIVE)
		    delta_start *= -1;
		  if (delta_start > 0)
		    fprintf(stdout, "positive side... ");
		  else
		    fprintf(stdout, "negative side... ");
		  status = man1d_build(&s, eval_cntr, delta_start, params);
		  printf(" s.fixpt = %f,%f\n", s.fixpt[0], s.fixpt[1]);
		  if (status >= 0){ /* edited by JE */
		     if(s1d==-1 && s.method==1){
		       status = man1d_iterate_noinv(&s, eval_cntr, params);
                     }
		     else status = man1d_iterate(&s, params);
                  }
		  if (status >= 0){
		    if(s1d==-1 && s.method==1){
		      status = man1d_grow_noinv(&s, eval_cntr, params);
		    }
		    else status = man1d_grow(&s, params);
		  }
            
                  if(s1d==-1 && s.method==1 && s.extend==1){
		    man1d_extend(&s, eval_cntr, delta_start, params);
		  }

                  fprintf(stdout, "arclength %f\n", s.mfld.arclength);
		  if (recording){
		    mfld_save(&(s.mfld), s.memory, params, 
			      s.prop_cntl.sys_color, s.prop_cntl.symbol, 
			      s.man_type);
		  }
		  empty_mfld(&(s.mfld));
		    
		  status = 0;
		  delta_start *= -1;
		}
	      }
	    }
	  }
	}
      }
    }
  }
 
  if (num == 0)
    system_mess_proc(1, "man1d_compute: No saddle points in memory.");

  release_fp(&fp, 0);
  release_man1d(&s);

  pm(EXEC, "Memory.Count", NULL);
  return;
}
