/*
 * file:   man1d_funct.c
 * author: Hinke Osinga - University of Bristol
 * date:   November 1997
 */


#include "utilities.h"
#include "math_utils.h"
#include "constants.h"
#include "complib.h"
#include "iterate.h"
#include "flow.h"
#include "man1d.h"


int F(double *y, double *x, struct Prop_DataS *iter_cntl, int *sw)
{
  int     status = 0, i, newt_iter, mc_iter;
  int     inv_failed = TRUE, v_dim = iter_cntl->ph_space_dim-1;
  int     explicit_switch = sw[0], approx_switch = sw[1];
  int     jacobian_switch = sw[2];
  double  *xhelp, *v, *h, **jac_matrix, tolx, tolf, fd_step, *work;

  jac_matrix = dmatrix(0, v_dim-1, 0, v_dim-1);
  xhelp = iter_cntl->workspace;                 /* x[ph_space_dim] */
  v = xhelp + iter_cntl->ph_space_dim;          /* v[ph_space_dim] */
  h = v + iter_cntl->ph_space_dim;            /* h[ph_space_dim-1] */
  work = h + v_dim;

  mc_iter = iter_cntl->panel_int_values[0];
  if (approx_switch) mc_iter = 1;
  newt_iter = iter_cntl->panel_int_values[1];
  fd_step = iter_cntl->panel_dp_values[0];
  tolx = iter_cntl->panel_dp_values[1];
  tolf = iter_cntl->panel_dp_values[2];

  for (i = 0; i < v_dim; i++)
    h[i] = fd_step;       /* for now, h(i) is the same for all i */

  if (iter_cntl->direction == FORWARD)
    /* forward iteration */
    iter_forw(iter_cntl->function, iter_cntl->iter_request, 
	      y, x, iter_cntl->parameters, iter_cntl->ph_space_dim, 
	      iter_cntl->time_step, work, iter_cntl->manifold);
  else{
    /* backward iteration */
    
    if (explicit_switch){
      /*   printf("Calculating stable manifold using explicit inverse.\n");*/
      iter_forw(iter_cntl->inverse, iter_cntl->iter_request,
		y, x, iter_cntl->parameters, iter_cntl->ph_space_dim,
		iter_cntl->time_step, work, iter_cntl->manifold);
    }
    else{
      inv_failed = TRUE;
      for(i = 0; i < mc_iter && inv_failed; i++){
	if (approx_switch){
	  /* printf("Calculating stable manifold using approximate inverse.\n");*/
	  iter_forw(iter_cntl->inverse, iter_cntl->iter_request, 
		    y, x, iter_cntl->parameters, iter_cntl->ph_space_dim, 
		    iter_cntl->time_step, work, iter_cntl->manifold); 
	}
	else{  
	  /* get monte-carlo guess */
	  /*printf("Calculating stable manifold using monte-carlo guess.\n");*/
	  y[v_dim] = x[v_dim] - (double)iter_cntl->iter_request; 
	  rnd_vector(v_dim, y, iter_cntl->min, iter_cntl->max);
	}
	inv_failed = 
	  implicit_iter(iter_cntl->function, iter_cntl->dfdx, 
			jacobian_switch, newt_iter, iter_cntl->iter_request, 
			jac_matrix, y, x, iter_cntl->parameters, 
			iter_cntl->ph_space_dim, iter_cntl->time_step, h, 
			tolx, tolf, iter_cntl->manifold, work, 
			iter_cntl->i_wkspace );
      }
       
      if (inv_failed){
	if (inv_failed == NO_CONVERGENCE){
	  system_mess_proc(1, "Implicit backwards iteration failed. ");
	  system_mess_proc(1, "No convergence within Newton's method.");
	}
	else if (inv_failed == SINGULAR){
	  system_mess_proc(1, "Implicit backwards iteration failed. ");
	  system_mess_proc(1, "Singular matrix in Newton's method");
	  free_dmatrix(jac_matrix, 0, v_dim-1, 0, v_dim-1 );
	  return (-2);
	}
      }
    }
  }

  for (i = 0; i < v_dim; i++){
    if (! finite(y[i]) || y[i] > iter_cntl->diverg_cutoff || 
	y[i] < -iter_cntl->diverg_cutoff){
      system_mess_proc(1, "Manifolds appears to diverge off to infinity! ");
      system_mess_proc(1, "Stop!"); 
      free_dmatrix(jac_matrix, 0, v_dim-1, 0, v_dim-1 );
      return (-2);
    }
  }

  free_dmatrix(jac_matrix, 0, v_dim-1, 0, v_dim-1 );

  return ( status );    
}

int FInv(double *x, double *y, struct Prop_DataS *iter_cntl, int *sw)
{
  int     status = 0, i, newt_iter, mc_iter;
  int     inv_failed = TRUE, v_dim = iter_cntl->ph_space_dim-1;
  int     explicit_switch = sw[0], approx_switch = sw[1];
  int     jacobian_switch = sw[2];
  double  *xhelp, *v, *h, **jac_matrix, tolx, tolf, fd_step, *work;

  jac_matrix = dmatrix(0, v_dim-1, 0, v_dim-1);
  xhelp = iter_cntl->workspace;                 /* x[ph_space_dim] */
  v = xhelp + iter_cntl->ph_space_dim;          /* v[ph_space_dim] */
  h = v + iter_cntl->ph_space_dim;            /* h[ph_space_dim-1] */
  work = h + v_dim;

  mc_iter = iter_cntl->panel_int_values[0];
  if (approx_switch) mc_iter = 1;
  newt_iter = iter_cntl->panel_int_values[1];
  fd_step = iter_cntl->panel_dp_values[0];
  tolx = iter_cntl->panel_dp_values[1];
  tolf = iter_cntl->panel_dp_values[2];

  for (i = 0; i < v_dim; i++)
    h[i] = fd_step;       /* for now, h(i) is the same for all i */

  if (iter_cntl->direction == FORWARD){
    /* backward iteration */
    if (explicit_switch)
      iter_forw(iter_cntl->inverse, iter_cntl->iter_request,
		x, y, iter_cntl->parameters, iter_cntl->ph_space_dim,
		iter_cntl->time_step, work, iter_cntl->manifold);
    else{
      inv_failed = TRUE;
      for(i = 0; i < mc_iter && inv_failed; i++){
	if (approx_switch)
	  iter_forw(iter_cntl->inverse, iter_cntl->iter_request, 
		    y, x, iter_cntl->parameters, iter_cntl->ph_space_dim, 
		    iter_cntl->time_step, work, iter_cntl->manifold); 
	else{  
	  /* get monte-carlo guess */
	  y[v_dim] = x[v_dim] - (double)iter_cntl->iter_request; 
	  rnd_vector(v_dim, y, iter_cntl->min, iter_cntl->max);
	}
	inv_failed = 
	  implicit_iter(iter_cntl->function, iter_cntl->dfdx, 
			jacobian_switch, newt_iter, iter_cntl->iter_request, 
			jac_matrix, y, x, iter_cntl->parameters, 
			iter_cntl->ph_space_dim, iter_cntl->time_step, h, 
			tolx, tolf, iter_cntl->manifold, work, 
			iter_cntl->i_wkspace );
      }
       
      if (inv_failed){
	if (inv_failed == NO_CONVERGENCE){
	  system_mess_proc(1, "Implicit backwards iteration failed. ");
	  system_mess_proc(1, "No convergence within Newton's method.");
	}
	else if (inv_failed == SINGULAR){
	  system_mess_proc(1, "Implicit backwards iteration failed. ");
	  system_mess_proc(1, "Singular matrix in Newton's method");
	  free_dmatrix(jac_matrix, 0, v_dim-1, 0, v_dim-1 );
	  return (-2);
	}
      }
    }
  }
  else {
    /* forward iteration */
    iter_forw(iter_cntl->function, iter_cntl->iter_request, 
	      x, y, iter_cntl->parameters, iter_cntl->ph_space_dim, 
	      iter_cntl->time_step, work, iter_cntl->manifold);
  }
  for (i = 0; i < v_dim; i++){
    if (! finite(y[i]) || y[i] > iter_cntl->diverg_cutoff || 
	y[i] < -iter_cntl->diverg_cutoff){
      system_mess_proc(1, "Manifolds appears to diverge off to infinity! ");
      system_mess_proc(1, "Stop!"); 
      free_dmatrix(jac_matrix, 0, v_dim-1, 0, v_dim-1 );
      return (-2);
    }
  }

  free_dmatrix(jac_matrix, 0, v_dim-1, 0, v_dim-1 );

  return ( status );    
}
