/*
 * file:   man1d_list.c
 * author: Hinke Osinga - University of Bristol
 * date:   November 1997
 */

/*
 * list_create(double *p, double *finvp, int dim)
 * list_free(List l, int dim)
 * mfld_load(ONED_MAN *mfld, int dim)
 * mfld_release(ONED_MAN *mfld)
 * add_to_mfld(ONED_MAN *mfld, double *p, double *finvp, List invleft)
 * empty_mfld(ONED_MAN *mfld)
 * mfld_save(ONED_MAN *mfld, memory m, double *params, int color,
 *	     int symbol, int mfld_type)
 */


#include <stdlib.h>
#include <stdio.h>

#include "memory.h"
#include "utilities.h"
#include "math_utils.h"
#include "plot.h"
#include "man1d_list.h"


List list_create(double *p, double *finvp, int dim)
{
  int      i;
  List l = (List) malloc((unsigned) sizeof(LIST_ITEM));

  l->next = NULL;
  l->prev = NULL;
  l->invleft = NULL;
  l->data = dvector(0, dim);
  l->inverse = dvector(0, dim);

  for (i = 0; i <= dim; i++){
    l->data[i] = p[i];
    l->inverse[i] = finvp[i];
  }

  return l;
}

void list_free(List l, int dim)
{
  free_dvector(l->data, 0, dim);
  free_dvector(l->inverse, 0, dim);
  free(l);
}

int  mfld_load(ONED_MAN *mfld, int dim)
{
  int  status = 0;

  mfld->size = 0;
  mfld->v_dim = dim;
  mfld->arclength = 0;

  return status;
}

void mfld_release(ONED_MAN *mfld)
{
  mfld->size = 0;
  mfld->v_dim = 0;
  mfld->arclength = 0;
}

int add_to_mfld(ONED_MAN *mfld, double *p, double *finvp, List invleft)
{
  int   status = 0;
  List  tail, new;

  new = list_create(p, finvp, mfld->v_dim);
  tail = mfld->tail;
  if (mfld->size > 0){
    tail->next = new;
    new->prev = tail;
    new->invleft = invleft;
  }
  else{
    /* first point is its own inverse */
    new->invleft = new;
    mfld->head = new;
  }
  mfld->tail = new;
  mfld->size++;
  return status;
}

void empty_mfld(ONED_MAN *mfld)
{
  List l, more;

  if (mfld->size == 0)
    return;
  l = mfld->tail;
  while (l != NULL){
    more = l->prev;
    list_free(l, mfld->v_dim);
    l = more;
  }
  mfld->size = 0;
}


void mfld_save(ONED_MAN *mfld, memory m, double *params, int color, 
	       int symbol, int mfld_type)
{
  int   c3[3], i, size;
  List  l;

  /* set up color stuff */
  c3[0] = -1;        /* force use of pick color */
  c3[1] = color;
  c3[2] = symbol;

  l = mfld->head;
  if (l == NULL){
    system_mess_proc(1, "mfld_save: No points to save!\n");
    return;
  }
  
  /* save funddom as a trajectory or set of trajectories */
  size = mfld->size;
  while (size > TRAJ_MAX){
    if (0 == memory_start_new_flow(m, 1, 0, 0, TRAJ_MAX, 0, 1)){
      /* go through each point in funddom */
      memory_add_point(m, NULL, params, c3, NULL, NULL, NULL, &mfld_type);
      for (i = 0; i < TRAJ_MAX; i++){
	memory_add_point(m, l->data, NULL, NULL, NULL, NULL, NULL, NULL);
	l = l->next;
      }
    }
    size -= TRAJ_MAX;
  }
  
  if (size > 0){
    if (0 == memory_start_new_flow(m, 1, 0, 0, size, 0, 1)){
      /* go through each point in funddom */
      memory_add_point(m, NULL, params, c3, NULL, NULL, NULL, &mfld_type);
      do{
	memory_add_point(m, l->data, NULL, NULL, NULL, NULL, NULL, NULL);
	l = l->next;
      } while(l != NULL);
    }
  }
}
