/*
 * file:   man1d_noinv_acc.c
 * author: James England - University of Bristol
 * date:   November 2002
 */

/* 
 * Find_Angle(double *p, double *pm1, int dim, struct Prop_DataS *iter_cntl)
 * Side_of_Line(double *Pi, double *Pim1, double *fq, int dim, 
 *              double *dotProd, struct Prop_DataS *iter_cntl)
 * Near_Line(double *Pi, *Pim1, *fq, int dim, struct Prop_DataS *iter_cntl)
 * Try_Point_noinv(int neg_eval, double delta_c, angle, *p, *p_test, 
 *		   double *fp, struct Prop_DataS *iter_cntl, int dim, int *sw)
 * Search_Circle(int neg_eval, double *pprev, double *pleft, double *pright, 
 *		 double *p, double delta_circle, int dim, double *pnew, 
 *		 double *invpnew, struct Prop_DataS *iter_cntl, int *sw, 
 *		 struct Man1D_DataS *s)
 * Check_Right_Section(List list_left, double *inv , int *rightsection, dim, 
 *	       int new, struct Prop_DataS *iter_cntl, struct Man1D_DataS *s)
 * Find_Point_alternative(int neg_eval, double *pnew, *invpnew, *p,  
 *			  double *prev, int dim, int new, List inv_item, 
 *			  double delta_circle, struct Prop_DataS *iter_cntl, 
 *			  int *sw, int *status, struct Man1D_DataS *s)
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "complib.h"
#include "utilities.h"
#include "math_utils.h"
#include "manifold.h"
#include "man1d_funct.h"
#include "man1d_list.h"
#include "plot.h"
#include "man1d.h"


double Find_Angle(double *p, double *pm1, int dim, struct Prop_DataS *iter_cntl)
{
    double length, theta;
    static int n=0;

    length = Norm(dim, p, pm1, iter_cntl->workspace, iter_cntl->manifold);
    theta = acos((p[1]-pm1[1])/length);
    if( (p[0]-pm1[0]) < 0 ) theta = 2*M_PI - theta;

    n++;
    return theta;
}

int Side_of_Line ( double *Pi, double *Pim1, double *fq, int dim, 
		   double *dotProd, struct Prop_DataS *iter_cntl)
{
    double *v = dvector(0, dim);
    double *w = dvector(0, dim);
    double *w1 = dvector(0, dim);
    double length;
    int i;

    for (i = 0; i < dim; i++){
      v[i] = fq[i]-Pi[i];
      w1[i] = Pim1[i]-Pi[i];
    }

    length = Norm(dim, Pi, Pim1, iter_cntl->workspace, iter_cntl->manifold);
   
    w[0] = w1[1]/length;
    w[1] = -1*w1[0]/length; 

    *dotProd = dot(dim, v, w);

    free_dvector(v, 0, dim);
    free_dvector(w, 0, dim);
    free_dvector(w1, 0, dim);

    if (*dotProd <= 0) {
      *dotProd = - *dotProd ;
      return 1;
    }
    else return 0;
}


int Near_Line (double *Pi, double *Pim1, double *fq, int dim, 
	       struct Prop_DataS *iter_cntl)
{
    double *v = dvector(0, dim);
    double *w  = dvector(0, dim);
    double dotProd;
    int i;

    for (i = 0; i <= dim; i++){
      v[i] = fq[i]-Pi[i];
      w[i] = Pim1[i]-Pi[i];
    }

    dotProd = dot(dim, v, w);

    free_dvector(v, 0, dim);
    free_dvector(w, 0, dim);

    if (dotProd > 0) return 1;
    else return 0;
}


int Try_Point_noinv(double delta_c, double angle, double *p, 
		    double *p_test, double *fp, struct Prop_DataS *iter_cntl, 
		    int dim, int *sw)
{
    double *temp = dvector(0, dim);
    double *fp2 = dvector(0, dim);
    int status=0;
    
    p_test[0] = (delta_c * sin(angle)) + p[0];
    p_test[1] = (delta_c * cos(angle)) + p[1];

    status = FInv(fp, p_test, iter_cntl, sw);
    
    free_dvector(fp2, 0, dim);
    free_dvector(temp,0, dim);

    return status;
}


int Search_Circle(double *pprev, double *pleft, double *pright, 
		  double *p, double delta_circle, int dim, double *pnew, 
		  double *invpnew, struct Prop_DataS *iter_cntl, int *sw, int init,
		  struct Man1D_DataS *s)
{
  
  int i, status=0;
  double *p_test = dvector(0, dim);
  double *p_left = dvector(0, dim);
  double *p_right = dvector(0, dim);
  double *Fp = dvector(0, dim);
  double *p_prev = dvector(0, dim);
  double theta, alpha, dotProd;
  int crossed, crossed_start, crossed_end, alpha_changed=0;
  double theta_try, start, end, zoom=0, err, ostart, oend;

  for (i=0; i < dim; i++){
	p_left[i] = pleft[i];
  	p_right[i] = pright[i];
        p_prev[i] = pprev[i];
  }

  theta = Find_Angle(p, p_prev, dim, iter_cntl);

  alpha = s->amax;

  start = theta - alpha;
  end = theta + alpha;
  ostart = start;
  oend = end;

  free_dvector(p_prev, 0, dim);

  do {

    /* --- checking whether p_start and p_end are on opposite sides of line ----*/

    status = Try_Point_noinv(delta_circle, start, p, p_test, Fp, 
			     iter_cntl, dim, sw);
    crossed_start = Side_of_Line(p_right, p_left, Fp, dim, &dotProd, iter_cntl);
    if( crossed_start == 1 ) {
    	start = start - 0.1;
        alpha_changed = 1;  
	if(start < ostart - 2*M_PI){
	  printf("Error - Start angle not found. p_i-1=%f,%f p_i=%f,%f fp=%f,%f\n",
		 p_left[0],p_left[1],p_right[0],p_right[1],Fp[0],Fp[1]);
	  return -2;
	}
    }

    status = Try_Point_noinv(delta_circle, end, p, p_test, Fp, 
			     iter_cntl, dim, sw);
    crossed_end = Side_of_Line(p_right, p_left, Fp, dim, &dotProd, iter_cntl);
    if( crossed_end == 0 ) {
	end = end + 0.1;
        alpha_changed = 1;
	if(end > oend+2*M_PI) {
	  printf("Error - End angle not found. p_i-1=%f,%f p_i%f,%f fp=%f,%f\n"
		 ,p_left[0],p_left[1],p_right[0],p_right[1],Fp[0],Fp[1]);
	  return -2;
	}
    }
    if( (crossed_start == 1 || crossed_end == 0) && init ==1) {
      return -3;
    }
    
    if(crossed_end == 0 && crossed_start == 1)  {
	  /* p_start and p_end on wrong sides of line   */
    }
  } while ((crossed_end == 0 || crossed_start == 1) && status >=0);

  /*if(alpha_changed==1) {
    printf("Warning - alpha search region increased. start alpha=%f, end alpha=%f (may not be used)\n", 
	   start-theta, end-theta);
  }*/

  /* -------------------------------------------------------------------------*/

  while(zoom < 60 && status>=0 ){
 
    theta_try = (end + start)/2;
    status = Try_Point_noinv(delta_circle, theta_try, p, p_test, 
			     Fp, iter_cntl, dim, sw);
 
    /* --- plot points that are searched and where they map ---- */

    /*point_plot(p_test, iter_cntl->parameters, 11, s->prop_cntl.symbol);
      point_plot(Fp, iter_cntl->parameters, 10, s->prop_cntl.symbol);*/

    /* --------------------------------------------------------- */

    crossed = Side_of_Line(p_right, p_left, Fp, dim, &dotProd, iter_cntl);
    if( crossed == 1 ) {
      end = theta_try;
    }
    else {
      start = theta_try;
    }

    if(dotProd < s->bierror){
      /* candidate for p_k+1 found. */
      for (i = 0; i < dim; i++){
	pnew[i] = p_test[i];
	invpnew[i] = Fp[i];
      }
      return status;
    }
    zoom++;
  }

  status = Try_Point_noinv(delta_circle, start, p, p_test, Fp, 
			   iter_cntl, dim, sw);
  err = abs(dotProd);

  fprintf(stdout, " Point on line not found! - closest taken P=%f,%f fp=%f,%f error=%f (May not be used)\n", 
	  p[0], p[1], p_test[0],p_test[1], err);

  for (i = 0; i < dim; i++){
    pnew[i] = p_test[i];
    invpnew[i] = Fp[i];
  }

  free_dvector(Fp, 0, dim);
  free_dvector(p_test, 0, dim);
  free_dvector(p_left, 0, dim);
  free_dvector(p_right, 0, dim);

  return status;
}

List Check_Right_Section( List list_left, double *inv , int *rightsection, int dim, 
			  int new, struct Prop_DataS *iter_cntl, struct Man1D_DataS *s)
{
    double tau1, tau2, ki, kim1;
    double *p_left = dvector(0, dim);  
    double *p_right = dvector(0, dim);
    int i, j;
    static int last=0;
  
    if (new==1) last=0;

    for (i = 0; i < dim; i++){
	p_left[i] = list_left->data[i];
	p_right[i] = list_left->next->data[i];
    }

    ki = Norm(dim, inv, p_right, iter_cntl->workspace, iter_cntl->manifold);
    kim1 = Norm(dim, inv, p_left, iter_cntl->workspace, iter_cntl->manifold);
    tau1 = ki/Norm(dim, p_left,p_right, iter_cntl->workspace, iter_cntl->manifold);
    tau2 = kim1/Norm(dim, p_left,p_right, iter_cntl->workspace, iter_cntl->manifold);

    free_dvector(p_left, 0, dim);
    free_dvector(p_right, 0, dim);

    if( tau1 <= 1  &&  tau2 <= 1) {
      /* point in correct section */
      *rightsection = 1;
      last=0;
      return list_left;
    }

    *rightsection=0;

    if( tau2 > 1 && ki < kim1){  
      /* Wrong section mapped to-->move forwards */
      if(last==-1){
	printf("Error - Oscillating between search regions. \n");       	
	*rightsection = -1;
	return list_left;
      }
      if(list_left->next->next==NULL){
	printf("Not mapped to last section - need to decrease delta_try\n");
	*rightsection = 3;
	return NULL;
      }

      last=1;
     
      for (j=0; j < s->num_extpt; j++){
	if( Norm(dim, p_right, s->finpt[j], iter_cntl->workspace, iter_cntl->manifold) 
	    < 0.000001){  /* computer accuracy */
	  /* last point was end of branch - skipping to new branch */
	   list_left = list_left->next;
	   printf("Jumped to check newly found piece of manifold.\n");
	}
      }

      return list_left->next;
    }

    else if( tau1 > 1){
      /* Wrong section mapped to <-- move back  */
      if(last==1){
	printf("Error - Oscillating between search regions.(trying to go back) \n");
	*rightsection = -1;
	return list_left;
      }
        
      last = -1;
      
      if(list_left->prev==NULL){
	printf("Error - No more sections to move back to.\n");
	*rightsection = -2;
	last=0;
	return list_left;
      }
      for (j=0; j < s->num_extpt; j++){
	if(Norm(dim, p_left, s->extpt[j], iter_cntl->workspace, iter_cntl->manifold)==0){  
	  list_left = list_left->prev;
	  printf("Jumped back to check on main manifold.\n");
	}
      }
      return list_left->prev;
    }

    else {
      printf("CRS - shouldn't get here! ki=%f, kim1=%f tau1=%f tau2=%f\n",
	     ki,kim1,tau1,tau2);
      *rightsection = -1;
    }

    last=0;
    return NULL;
}


List Find_Point_alternative(double *pnew, double *invpnew, double *p,  
			    double *prev, int dim, int new, List inv_item, 
			    double delta_circle, struct Prop_DataS *iter_cntl, 
			    int *sw, int *status, int init, struct Man1D_DataS *s)
{
  double   *pleft = dvector(0, dim);
  double   *pright = dvector(0, dim);
  double  half_dmin;
  int 	       right_section, i, n=0;
  int          nw;
  List         list_elem;

  nw = new;
  list_elem = inv_item;

  for (i = 0; i < dim; i++){
    pleft[i] = inv_item->data[i];
    pright[i] = inv_item->next->data[i];
  }
  
  *status=-2;
  half_dmin = s->dmin / 2;
  
  do{
      *status = Search_Circle(prev, pleft, pright, p, delta_circle, 
			      dim, pnew, invpnew, iter_cntl, sw, init, s);

      if(*status >= 0) {
	list_elem = Check_Right_Section( list_elem, invpnew, &right_section, 
					 dim, nw, iter_cntl, s);
        nw = 0;
	if(list_elem != NULL){
	  for (i = 0; i < dim; i++){
	    pleft[i] = list_elem->data[i];
	    pright[i] = list_elem->next->data[i];
	  }
	}
        if(right_section ==-1) {
          printf("Oscillating between regions, try decreasing delta\n");
          if(delta_circle > s->dmin) {
	    delta_circle = delta_circle/2;
	    *status=0;
            right_section = 0;
            nw = 1;
	    if(delta_circle < half_dmin) delta_circle = half_dmin;

	  }
	  else {
	    printf("Delta is at minimum and still oscillating. Stop.\n");
	    *status = -1;
	  }
	}
      }
      if(right_section==-2 || (*status==-2 && list_elem->prev==NULL)){
	s->prop_cntl.iter_request++;
	if(s->prop_cntl.iter_request == s->max_iter+1) s->prop_cntl.iter_request=1;
	
	right_section=0;
	if(n < s->max_iter-1) {
	  *status=0; /* Haven't tried all iterates */
	  printf("Swapping to iterate %d.\n", s->prop_cntl.iter_request);
	}
	else *status=-2;
	n++;
      }
      
  } while( right_section == 0 && *status>=0 );

  free_dvector(pleft, 0, dim);
  free_dvector(pright, 0, dim);
  return list_elem;
}
